"""
https://github.com/siemens/vanilc/blob/master/3rdParty/boost/math/special_functions/lanczos.hpp
https://github.com/siemens/vanilc/blob/master/3rdParty/boost/math/special_functions/gamma.hpp
"""
import decimal
import math
import sys
from decimal import Decimal, localcontext
from typing import Union

if sys.version_info < (3, 9):
    from typing import Reversible
else:
    from collections.abc import Reversible

G = Decimal("10.90051099999999983936049829935654997826")

NUMERATORS = (*[
    Decimal(x)
    for x in (
        "38474670393.31776828316099004518914832218",
        "36857665043.51950660081971227404959150474",
        "15889202453.72942008945006665994637853242",
        "4059208354.298834770194507810788393801607",
        "680547661.1834733286087695557084801366446",
        "78239755.00312005289816041245285376206263",
        "6246580.776401795264013335510453568106366",
        "341986.3488721347032223777872763188768288",
        "12287.19451182455120096222044424100527629",
        "261.6140441641668190791708576058805625502",
        "2.506628274631000502415573855452633787834"
    )
],)

DENOMINATORS = (
    0,
    362880,
    1026576,
    1172700,
    723680,
    269325,
    63273,
    9450,
    870,
    45,
    1,
)

EXP_NUMERATORS = (*[
    Decimal(x)
    for x in (
        "709811.662581657956893540610814842699825",
        "679979.847415722640161734319823103390728",
        "293136.785721159725251629480984140341656",
        "74887.5403291467179935942448101441897121",
        "12555.29058241386295096255111537516768137",
        "1443.42992444170669746078056942194198252",
        "115.2419459613734722083208906727972935065",
        "6.30923920573262762719523981992008976989",
        "0.2266840463022436475495508977579735223818",
        "0.004826466289237661857584712046231435101741",
        "0.00004624429436045378766270459638520555557321"
    )
],)

def evaluate_rational(
    numerator: Reversible[Union[Decimal, int]],
    denominator: Reversible[Union[Decimal, int]],
    x: Decimal,
) -> Decimal:
    s1 = 0
    s2 = 0
    if x > 1:
        x = 1 / x
    else:
        numerator = reversed(numerator)
        denominator = reversed(denominator)
    for coef in numerator:
        s1 *= x
        s1 += coef
    for coef in denominator:
        s2 *= x
        s2 += coef
    return s1 / s2

def digamma_P_lanczos(x: Decimal) -> Decimal:
    derivative = [i * x for i, x in enumerate(EXP_NUMERATORS[1:], 1)]
    derivative.append(0)
    return evaluate_rational(derivative, EXP_NUMERATORS, x)

def digamma_Q_lanczos(x: Decimal) -> Decimal:
    derivative = [i * x for i, x in enumerate(DENOMINATORS[1:], 1)]
    derivative.append(0)
    return evaluate_rational(derivative, DENOMINATORS, x)

def gamma_lanczos(x: Decimal) -> Decimal:
    return evaluate_rational(NUMERATORS, DENOMINATORS, x)

def lgamma_exp_lanczos(x: Decimal) -> Decimal:
    return evaluate_rational(EXP_NUMERATORS, DENOMINATORS, x)

def digamma_precise(x: float) -> Decimal:
    from .imath import _BIG_PI, cos_precise, sin_precise
    with localcontext() as ctx:
        ctx.prec = 72
        if x > 0.0 and math.isinf(x):
            return Decimal("Infinity")
        elif x > 0.0:
            d = Decimal(x)
        elif math.isinf(x):
            assert False, "digamma_precise should not accept negative infinity"
        elif x.is_integer():
            assert False, "digamma_precise should not accept non-positive integers"
        else:
            d = 1 - Decimal(x)
        if d >= 1e4:
            d2 = d ** -2
            result = -d2 / 12
            result += Decimal(691) / 32760
            result *= d2
            result -= Decimal(1) / 132
            result *= d2
            result += Decimal(1) / 240
            result *= d2
            result -= Decimal(1) / 252
            result *= d2
            result += Decimal(1) / 120
            result *= d2
            result -= Decimal(1) / 12
            result /= d
            result -= Decimal("0.5")
            result /= d
            result += d.ln()
        else:
            d1 = d % 1
            dgh = d1 + G + Decimal("0.5")
            result = (
                (d1 + Decimal("0.5")) / dgh
                + dgh.ln()
                + digamma_P_lanczos(d1 + 1)
                - digamma_Q_lanczos(d1 + 1)
                - 1
            )
            dx = 0
            while d - d1 > 1.5:
                d1 += 1
                dx += 1 / d1
            while d - d1 < 0.5:
                dx -= 1 / d1
                d1 -= 1
            result += dx
        if x <= 0.0:
            theta = x % 1.0
            if theta > 0.5:
                theta = -((-x) % 1.0)
            c = cos_precise(_BIG_PI * Decimal(theta))
            s = sin_precise(_BIG_PI * Decimal(theta))
            result -= _BIG_PI * c / s
        return result

def gamma_precise(x: float) -> Decimal:
    from .imath import _BIG_PI, sin_precise
    with localcontext() as ctx:
        ctx.prec = 72
        if x >= 179.0:
            return Decimal("Infinity")
        elif x > 0.0 and x.is_integer():
            result = Decimal(1)
            temp = 1
            for i in range(2, round(x)):
                temp *= i
                if temp > 1e30:
                    result *= temp
                    temp = 1
            return result * temp
        elif x > 0.0:
            d = Decimal(x)
        elif math.isinf(x):
            assert False, "gamma_precise should not accept negative infinity"
        elif x.is_integer():
            assert False, "gamma_precise should not accept non-positive integers"
        else:
            d = 1 - Decimal(x)
        result = None
        if d >= 10.0:
            d2 = d ** -2
            s = d2 / 156
            s -= Decimal(691) / 360360
            s *= d2
            s += Decimal(1) / 1188
            s *= d2
            s -= Decimal(1) / 1680
            s *= d2
            s += Decimal(1) / 1260
            s *= d2
            s -= Decimal(1) / 360
            s *= d2
            s += Decimal(1) / 12
            s /= d
            try:
                result = (2 * _BIG_PI / d).sqrt() * (d * Decimal(-1).exp()) ** d * s.exp()
            except decimal.Overflow:
                result = Decimal("Infinity")
        else:
            d1 = d % 1
            result = ((d1 + G + Decimal("0.5")) * Decimal(-1).exp()) ** (d1 + Decimal("0.5")) * gamma_lanczos(d1 + 1)
            dx = 1
            while d - d1 > 1.5:
                d1 += 1
                dx *= d1
            while d - d1 < 0.5:
                dx /= d1
                d1 -= 1
            result *= dx
        if x <= 0.0:
            theta = x % 2.0
            if theta > 1.0:
                theta = -((-x) % 2.0)
            result = _BIG_PI / (
                sin_precise(Decimal(theta) * _BIG_PI)
                * result
            )
        return result

def lgamma_precise(x: float) -> Decimal:
    from .imath import _BIG_PI, sin_precise
    with localcontext() as ctx:
        ctx.prec = 72
        if x > 0.0 and math.isinf(x):
            return math.inf
        elif 0.0 < x < 3e4 and x.is_integer():
            result = Decimal(0)
            temp = 1
            for i in range(2, round(x)):
                temp *= i
                if temp > 1e30:
                    result += Decimal(temp).ln()
                    temp = 1
            return result + Decimal(temp).ln()
        elif x > 0.0:
            d = Decimal(x)
        elif math.isinf(x):
            assert False, "lgamma_precise should not accept negative infinity"
        elif x.is_integer():
            assert False, "lgamma_precise should not accept non-positive integers"
        else:
            d = 1 - Decimal(x)
        if -9.0 < x < 10.0:
            d1 = d % 1
            result = ((d1 + Decimal("0.5") + G).ln() - 1) * (d1 + Decimal("0.5")) + lgamma_exp_lanczos(d1 + 1).ln()
            dx = 0
            while d - d1 > 1.5:
                d1 += 1
                dx += d1.ln()
            while d - d1 < 0.5:
                dx -= d1.ln()
                d1 -= 1
            result += dx
        else:
            d2 = d ** -2
            s = d2 / 156
            s -= Decimal(691) / 360360
            s *= d2
            s += Decimal(1) / 1188
            s *= d2
            s -= Decimal(1) / 1680
            s *= d2
            s += Decimal(1) / 1260
            s *= d2
            s -= Decimal(1) / 360
            s *= d2
            s += Decimal(1) / 12
            s /= d
            result = (
                s
                + (2 * _BIG_PI).ln() / 2
                - d.ln() / 2
                + d * (d.ln() - 1)
            )
        if x <= 0.0:
            theta = abs(x) % 2.0
            if theta > 1.0:
                theta = (-abs(x)) % 2.0
            result = (
                _BIG_PI.ln()
                - sin_precise(Decimal(theta) * _BIG_PI).ln()
                - result
            )
        return result



import ctypes
import math
import sys
from decimal import Decimal
from typing import SupportsFloat

if sys.version_info < (3, 9):
    from typing import Callable, Iterator, List, Tuple
else:
    from builtins import list as List, tuple as Tuple
    from collections.abc import Callable, Iterator

SUBNORMAL = math.ldexp(1.0, -1022)
NEAR_ZERO = math.ldexp(1.0, -1074)

nextafter: Callable[[float, float], float]

try:
    if sys.version_info >= (3, 9):
        from math import nextafter
    elif sys.platform == "linux" or sys.platform == "linux2":
        nextafter = ctypes.cdll.LoadLibrary("libm.so.6").nextafter
        nextafter.restype = ctypes.c_double
        nextafter.argtypes = [ctypes.c_double, ctype.c_double]
    elif sys.platform == "darwin":
        nextafter = ctypes.cdll.LoadLibrary("libSystem.dylib").nextafter
        nextafter.restype = ctypes.c_double
        nextafter.argtypes = [ctypes.c_double, ctype.c_double]
    elif sys.platform == "win32":
        nextafter = ctypes.cdll.LoadLibrary("libSystem.dylib")._nextafter
        nextafter.restype = ctypes.c_double
        nextafter.argtypes = [ctypes.c_double, ctype.c_double]
    else:
        from math import nextafter
except:
    def nextafter(x: float, y: float) -> float:
        if isinstance(x, SupportsFloat):
            x = float(x)
        if isinstance(y, SupportsFloat):
            y = float(y)
        if x == y:
            return x
        elif math.isnan(x) or math.isnan(y):
            return x + y
        elif math.isinf(x):
            return x
        elif -SUBNORMAL < x < SUBNORMAL:
            if x < y:
                return x - NEAR_ZERO
            else:
                return x + NEAR_ZERO
        mantissa, exponent = math.frexp(x)
        if x < y:
            mantissa += 0.5 * sys.float_info.epsilon
        else:
            mantissa -= 0.5 * sys.float_info.epsilon
        return math.ldexp(mantissa, exponent)

from .typing import SupportsRichFloat

def split_bits(n: int) -> Iterator[int]:
    if n > 0:
        for i in range(0, n.bit_length(), 53):
            yield (n % (1 << 53)) << i
            n >>= 53
    else:
        n = -n
        for i in range(0, n.bit_length(), 53):
            yield -((n % (1 << 53)) << i)
            n >>= 53

def split_precision(x: float) -> Tuple[float, float, float]:
    x_mantissa, x_exponent = math.frexp(x)
    sqrt_eps = math.sqrt(sys.float_info.epsilon)
    x_small = math.remainder(x_mantissa, sqrt_eps)
    x_large = x_mantissa - x_small
    return (x_exponent, x_small, x_large)

def float_split(x: SupportsRichFloat) -> Tuple[float, float]:
    y = float(x)
    if x < y:
        L = nextafter(y, -math.inf)
        U = y
    elif x > y:
        L = y
        U = nextafter(y, math.inf)
    else:
        L = U = y
    return (L, U)

def float_down(x: SupportsRichFloat) -> float:
    y = float(x)
    if x < y:
        return nextafter(y, -math.inf)
    else:
        return y

def float_up(x: SupportsRichFloat) -> float:
    y = float(x)
    if x > y:
        return nextafter(y, math.inf)
    else:
        return y

def multi_add(*args: float) -> List[float]:
    partials = []
    remaining = sorted(args)
    for _ in range(len(args)):
        if len(partials) == 0 or partials[-1] < 0:
            partials_add(partials, remaining.pop())
        else:
            partials_add(partials, remaining.pop(0))
    return partials

def partials_add(partials: List[float], x: float) -> List[float]:
    i = 0
    for y in partials:
        if abs(x) < abs(y):
            x, y = y, x
        total = x + y
        if math.isinf(total):
            partials[:] = [total]
            return partials
        error = y - (total - x)
        if error != 0.0:
            partials[i] = error
            i += 1
        x = total
    partials[i:] = [x]
    if len(partials) > 1 and x == 0.0:
        del partials[-1]
    return partials

def partials_times(partials: List[float], x: float) -> List[float]:
    temp = []
    for y in partials:
        exponent, mantissas = mul_precise(x, y)
        for z in mantissas:
            partials_add(temp, math.ldexp(z, exponent))
    partials[:] = temp
    return partials

def add_precise(x: float, y: float) -> List[float]:
    if abs(x) < abs(y):
        x, y = y, x
    total = x + y
    error = y - (total - x)
    return [error, total]

def add_down(x: float, y: float) -> float:
    if isinstance(y, int):
        partials = multi_add(x, *[float(n) for n in split_bits(y)])
    else:
        partials = add_precise(x, y)
    if len(partials) > 1 and partials[-2] < 0.0:
        return nextafter(partials[-1], -math.inf)
    else:
        return partials[-1]

def add_up(x: float, y: float) -> float:
    if isinstance(y, int):
        partials = multi_add(x, *[float(n) for n in split_bits(y)])
    else:
        partials = add_precise(x, y)
    if len(partials) > 1 and partials[-2] > 0.0:
        return nextafter(partials[-1], math.inf)
    else:
        return partials[-1]

def sub_down(x: float, y: float) -> float:
    return add_down(x, -y)

def sub_up(x: float, y: float) -> float:
    return add_up(x, -y)

def mul_precise(x: float, y: float) -> Tuple[float, List[float]]:
    if math.isinf(x) or math.isinf(y) or math.isnan(x) or math.isnan(y):
        return [x * y]
    x_exponent, *xs = split_precision(x)
    y_exponent, *ys = split_precision(y)
    exponent = x_exponent + y_exponent
    return (exponent, multi_add(*[xi * yi for xi in xs for yi in ys]))

def mul_down(x: float, y: float) -> float:
    if math.isinf(x) or math.isinf(y) or x * y == -math.inf:
        return x * y
    elif x * y == math.inf:
        return sys.float_info.max
    exponent, partials = mul_precise(x, y)
    if len(partials) == 1 or partials[-2] >= 0.0:
        z = math.ldexp(partials[-1], exponent)
    else:
        z = math.ldexp(nextafter(partials[-1], -math.inf), exponent)
    z_mantissa, z_exponent = math.frexp(z)
    if z_mantissa <= math.ldexp(partials[-1], exponent - z_exponent):
        return z
    else:
        return nextafter(z, -math.inf)

def mul_up(x: float, y: float) -> float:
    if math.isinf(x) or math.isinf(y) or x * y == math.inf:
        return x * y
    elif x * y == -math.inf:
        return -sys.float_info.max
    exponent, partials = mul_precise(x, y)
    if len(partials) == 1 or partials[-2] <= 0.0:
        z = math.ldexp(partials[-1], exponent)
    else:
        z = math.ldexp(nextafter(partials[-1], math.inf), exponent)
    z_mantissa, z_exponent = math.frexp(z)
    if z_mantissa >= math.ldexp(partials[-1], exponent - z_exponent):
        return z
    else:
        return nextafter(z, math.inf)

def div_down(x: float, y: float) -> float:
    if y == 0.0:
        return -math.inf
    elif math.isinf(x) and math.isinf(y):
        if x < 0.0 < y or y < 0.0 < x:
            return -math.inf
        else:
            return 0.0
    quotient = x / y
    if quotient == math.inf and not math.isinf(x):
        return nextafter(math.inf, 0.0)
    elif math.isinf(quotient):
        return quotient
    elif quotient != 0.0:
        exponent, partials = mul_precise(quotient, y)
        p1 = partials[-1] if exponent < 0 else math.ldexp(partials[-1], exponent)
        x1 = math.ldexp(x, -exponent) if exponent < 0 else x
        if y > 0:
            if p1 > x1:
                return nextafter(quotient, -math.inf)
            elif p1 < x1 or len(partials) == 1 or partials[-2] < 0.0:
                return quotient
            else:
                return nextafter(quotient, -math.inf)
        else:
            if p1 < x1:
                return nextafter(quotient, -math.inf)
            elif p1 > x1 or len(partials) == 1 or partials[-2] > 0.0:
                return quotient
            else:
                return nextafter(quotient, -math.inf)
    elif x == 0.0 or math.isinf(y) or not (x < 0.0 < y or y < 0.0 < x):
        return quotient
    else:
        return nextafter(0.0, -math.inf)

def div_up(x: float, y: float) -> float:
    if y == 0.0:
        return math.inf
    elif math.isinf(x) and math.isinf(y):
        if x < 0.0 < y or y < 0.0 < x:
            return 0.0
        else:
            return math.inf
    quotient = x / y
    if quotient == -math.inf and not math.isinf(x):
        return nextafter(-math.inf, 0.0)
    elif math.isinf(quotient):
        return quotient
    elif quotient != 0.0:
        exponent, partials = mul_precise(quotient, y)
        p1 = partials[-1] if exponent < 0 else math.ldexp(partials[-1], exponent)
        x1 = math.ldexp(x, -exponent) if exponent < 0 else x
        if y > 0:
            if p1 < x1:
                return nextafter(quotient, math.inf)
            elif p1 > x1 or len(partials) == 1 or partials[-2] > 0.0:
                return quotient
            else:
                return nextafter(quotient, math.inf)
        else:
            if p1 > x1:
                return nextafter(quotient, math.inf)
            elif partials[-1] < x or len(partials) == 1 or partials[-2] < 0.0:
                return quotient
            else:
                return nextafter(quotient, math.inf)
    elif x == 0.0 or math.isinf(y) or x < 0.0 < y or y < 0.0 < x:
        return quotient
    else:
        return nextafter(0.0, math.inf)

def pow_down(x: float, y: float) -> float:
    try:
        result = math.pow(x, y)
    except OverflowError:
        return nextafter(x ** (y % 2) * math.inf, -math.inf)
    except ValueError:
        if y == round(y) and round(y) % 2 == 1:
            return -math.inf
        else:
            return math.inf
    if (
        not 0.0 != abs(x) != 1.0
        or math.isinf(x)
        or math.isnan(x)
        or math.isinf(y)
        or math.isnan(y)
    ):
        return result
    elif Decimal(x) ** Decimal(y) < result:
        return nextafter(result, -math.inf)
    else:
        return result

def pow_up(x: float, y: float) -> float:
    try:
        result = math.pow(x, y)
    except OverflowError:
        return nextafter(x ** (y % 2) * math.inf, math.inf)
    except ValueError:
        return math.inf
    if (
        not 0.0 != abs(x) != 1.0
        or math.isinf(x)
        or math.isnan(x)
        or math.isinf(y)
        or math.isnan(y)
    ):
        return result
    elif Decimal(x) ** Decimal(y) > result:
        return nextafter(result, math.inf)
    else:
        return result



import sys
from typing import Optional

if sys.version_info < (3, 9):
    from typing import Callable, Iterator
else:
    from collections.abc import Callable, Iterator

from ._src.interval import Interval
from ._src import root_all as _root_all

__all__ = ["bisect", "newton"]

FLOAT_EPSILON: float = 2 * sys.float_info.epsilon
FLOAT_MIN: float = sys.float_info.min

def bisect(
    f: Callable[[Interval], Interval],
    /,
    interval: Interval = Interval((-float("inf"), float("inf"))),
    *,
    x: Optional[float] = None,
    abs_err: float = 0.0,
    rel_err: float = 32 * FLOAT_EPSILON,
    abs_tol: Optional[float] = None,
    rel_tol: Optional[float] = None,
) -> Interval:
    abs_err = max(float(abs_err), 32 * FLOAT_MIN)
    if abs_tol is None:
        abs_tol = min(4.0625, abs(0.1 * interval.minimum - 0.1 * interval.maximum))
    rel_err = max(float(rel_err), 32 * FLOAT_EPSILON)
    abs_tol = max(float(abs_tol), abs_err)
    if rel_tol is None:
        rel_tol = rel_err / (1024 * FLOAT_EPSILON)
    rel_tol = max(float(rel_tol), rel_err)
    rel_err = min(rel_err, 0.5)
    rel_tol = min(rel_tol, 0.5)
    endpoints = []
    for solution in _root_all.bisect(f, interval, x, abs_err, rel_err, abs_tol, rel_tol):
        if (
            len(endpoints) == 0
            or solution.minimum - endpoints[-1] > abs_err + rel_err * abs(Interval((endpoints[-1], solution.minimum))).minimum
        ):
            endpoints.append(solution.minimum)
            endpoints.append(solution.maximum)
        else:
            endpoints[-1] = solution.maximum
    iterator = iter(endpoints)
    return Interval(*zip(iterator, iterator))

def newton(
    f: Callable[[Interval], Interval],
    fprime: Callable[[Interval], Interval],
    /,
    interval: Interval = Interval((-float("inf"), float("inf"))),
    *,
    x: Optional[float] = None,
    abs_err: float = 0.0,
    rel_err: float = 32 * FLOAT_EPSILON,
    abs_tol: Optional[float] = None,
    rel_tol: Optional[float] = None,
) -> Interval:
    abs_err = max(float(abs_err), 32 * FLOAT_MIN)
    if abs_tol is None:
        abs_tol = min(4.0625, abs(0.1 * interval.minimum - 0.1 * interval.maximum))
    rel_err = max(float(rel_err), 32 * FLOAT_EPSILON)
    abs_tol = max(float(abs_tol), abs_err)
    if rel_tol is None:
        rel_tol = rel_err / (1024 * FLOAT_EPSILON)
    rel_tol = max(float(rel_tol), rel_err)
    rel_err = min(rel_err, 0.5)
    rel_tol = min(rel_tol, 0.5)
    endpoints = []
    for solution in _root_all.newton(f, fprime, interval, x, abs_err, rel_err, abs_tol, rel_tol):
        if (
            len(endpoints) == 0
            or solution.minimum - endpoints[-1] > abs_err + rel_err * abs(Interval((endpoints[-1], solution.minimum))).minimum
        ):
            endpoints.append(solution.minimum)
            endpoints.append(solution.maximum)
        else:
            endpoints[-1] = solution.maximum
    iterator = iter(endpoints)
    return Interval(*zip(iterator, iterator))
